package gcr.mmm.ui;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Point;
import java.io.IOException;

import javax.swing.JFrame;

import edu.berkeley.guir.prefuse.Display;
import edu.berkeley.guir.prefuse.ItemRegistry;
import edu.berkeley.guir.prefuse.action.RepaintAction;
import edu.berkeley.guir.prefuse.action.filter.GraphFilter;
import edu.berkeley.guir.prefuse.activity.ActionList;
import edu.berkeley.guir.prefuse.graph.DefaultNode;
import edu.berkeley.guir.prefuse.graph.Graph;
import edu.berkeley.guir.prefuse.graph.io.XMLGraphReader;
import edu.berkeley.guir.prefuse.render.DefaultEdgeRenderer;
import edu.berkeley.guir.prefuse.render.DefaultRendererFactory;
import edu.berkeley.guir.prefuse.render.ShapeRenderer;
import edu.berkeley.guir.prefuse.render.TextImageItemRenderer;
import edu.berkeley.guir.prefusex.controls.DragControl;
import edu.berkeley.guir.prefusex.force.CircularWallForce;
import edu.berkeley.guir.prefusex.force.DragForce;
import edu.berkeley.guir.prefusex.force.ForceSimulator;
import edu.berkeley.guir.prefusex.force.NBodyForce;
import edu.berkeley.guir.prefusex.force.SpringForce;
import edu.berkeley.guir.prefusex.layout.ForceDirectedLayout;

/**
 * @author Benjamin Hill
 * MMM2 Project
 * BSD Licence
 */
public class MetaTagger extends JFrame {

    private static final String XML_GRAPH = "etc/test.xml";

    /**
     * Comment for <code>serialVersionUID</code>
     */
    private static final long serialVersionUID = 7610805399823885598L;

    final int WIDTH = 800;

    final int HEIGHT = 600;

    private ItemRegistry registry;

    private Graph graph;

    private ActionList filter, forces;

    VisibleCircularWallForce vcwf = new VisibleCircularWallForce(-3.0f,
            WIDTH / 2, HEIGHT / 2, 100);

    /**
     * Prefuse based display using photo nearness
     */
    public MetaTagger() {
        super("SIMS");
        try {
            this.init();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    private void init() throws IOException {
        XMLGraphReader gr = new XMLGraphReader();
        gr.setNodeType(DefaultNode.class);

        this.graph = gr.loadGraph(XML_GRAPH);

        this.registry = new ItemRegistry(this.graph);

        // set up a renderer such that nodes show text labels
        TextImageItemRenderer nodeRenderer = new TextImageItemRenderer();
        nodeRenderer.setRenderType(ShapeRenderer.RENDER_TYPE_NONE);
        DefaultEdgeRenderer edgeRenderer = new DefaultEdgeRenderer();
        edgeRenderer.setWidth(2);

        // create a new renderer factory
        this.registry.setRendererFactory(new DefaultRendererFactory(
                nodeRenderer, edgeRenderer, null));

        // create a new display component to show the data
        Display display = new BackgroundDisplay(this.registry);
        display.setSize(WIDTH, HEIGHT);

        // let users drag nodes around on screen
        boolean repaint = false;
        boolean fixOnMouseOver = true;
        display.addControlListener(new DragControl(repaint, fixOnMouseOver));

        // create a new action list that
        // (a) filters visual representations from the original graph
        this.filter = new ActionList(this.registry); // Rerun the filter!
        this.filter.add(new GraphFilter());

        // run this action list for an infinite duration (duration value
        // equals -1), re-executing the action list every 15 milliseconds
        this.forces = new ActionList(this.registry, -1, 15);

        ForceDirectedLayout fdl = new ForceDirectedLayout(
                new MyForceSimulator(), true);
        this.forces.add(fdl);

        this.forces.add(new RepaintAction());

        // set up this JFrame
        setDefaultCloseOperation(EXIT_ON_CLOSE);
        getContentPane().add(display);
        pack();
        setVisible(true);

        // now execute the action lists to visualize the graph
        this.filter.runNow(); // filter runs once
        this.forces.runNow(); // forces re-runs every 15ms, indefinitely

    }

    /**
     * Renders a circle on the background
     * 
     * @author Benjamin
     * 
     */
    class BackgroundDisplay extends Display {
        /**
         * @param ir
         */
        public BackgroundDisplay(ItemRegistry ir) {
            super(ir);
        }

        /**
         * @param g
         * @see edu.berkeley.guir.prefuse.Display#prePaint(java.awt.Graphics2D)
         */
        protected void prePaint(Graphics2D g) {
            vcwf.paint(g);
        }
    }

    /**
     * Gravitates towards the center
     * 
     * @author Benjamin Hill
     * 
     */
    class MyForceSimulator extends ForceSimulator {
        MyForceSimulator() {
            this.addForce(new NBodyForce());
            this.addForce(new SpringForce());
            this.addForce(new DragForce());
            // [drag], x, y, r
            this.addForce(vcwf);
        }
    }

    class VisibleCircularWallForce extends CircularWallForce {
        int x, y;

        float r;

        /**
         * Visisble circular wall force - see what you are doing.
         * 
         * @param gravConst
         * @param x
         * @param y
         * @param r
         */
        public VisibleCircularWallForce(float gravConst, int x, int y, int r) {
            super(gravConst, x, y, r);
            this.x = x;
            this.y = y;
            this.r = r;
        }

        void paint(Graphics2D g) {
            g.setColor(Color.red);
            // x y w h
            g.setStroke(new BasicStroke(5.0f));
            g.drawOval((int) (this.x - this.r), (int) (this.y - this.r),
                    (int) this.r * 2, (int) this.r * 2);
        }

        public Point getLocation() {
            return new Point(this.x, this.y);
        }
    }

    /**
     * @param argv
     */
    public static void main(String[] argv) {
        new MetaTagger();
    }
}