package gcr.mmm2.model;

import java.io.File;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;

import com.drew.metadata.*;
import com.drew.imaging.jpeg.*;

/**
 * Represents EXIF and other metadata for an image.
 * 
 * @author Mike Wooldridge
 * 
 */
public class JpegMetadata {
	
	private Metadata metadata;
	private File jpegFile;
	private HashMap jpegTags = new HashMap();
	private ArrayList jpegTagsList = new ArrayList();
	
	private String sep = "%%%"; // Separator for creating directory/name hash map keys
	
	private final String DIRECTORY_EXIF = "Exif";
	private final String DIRECTORY_JPEG = "Jpeg";
	private final String NAME_DATETIME = "Date/Time";
    private final String[] NAME_POSSDATETIME = {"Date/Time","Date/Time Original"};
	private final String NAME_WIDTH = "Image Width";
	private final String NAME_HEIGHT = "Image Height";
	
	/**
	 * Takes JPEG file and loads metadata items into hash map. 
	 * Each item includes a directory, name, and description.
	 * Uses EXIF-processing classes from com.drew.metadata.*.
	 * 
	 * @param File path and name of JPEG image file. Only works for JPEG files.
	 */
	public JpegMetadata(String filePathAndName) {
		try {
			jpegFile = new File(filePathAndName);
			metadata = JpegMetadataReader.readMetadata(jpegFile);
			// iterate through metadata directories
			Iterator directories = metadata.getDirectoryIterator();
			while (directories.hasNext()) {
			    Directory directory = (Directory)directories.next();
			    // iterate through tags
			    Iterator tags = directory.getTagIterator();
			    while (tags.hasNext()) {
			        Tag tag = (Tag)tags.next();
			        System.out.println(tag);
			        JpegMetadataTag jpegTag = new JpegMetadataTag(directory.getName(), tag.getTagName(), tag.getDescription());
			        // Use name directory and name to build hash map key
			        String key = directory.getName() + sep + tag.getTagName();
			        jpegTags.put(key, jpegTag);
			        jpegTagsList.add(jpegTag);
			    }
			}
		} catch (JpegProcessingException e) {
			e.printStackTrace();
		} catch (MetadataException e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Get the description value of a metadata item given a directory and name.
	 * 
	 * @param directory (example: "Exif")
	 * @param name (example: "Date/Time")
	 * @return description
	 * 
	 */
	public String getDescription(String directory, String name) {
		String k = directory + sep + name;
		if (jpegTags.containsKey(k)) {
			JpegMetadataTag jpegTag = (JpegMetadataTag) jpegTags.get(k);
			return jpegTag.getDescription();	
		}
		return null;
	}
		
	/**
	 * Get EXIF Date/Time description (time of capture of image)
	 * 
	 * @return Date/Time as Date object
	 * 
	 */
	public Date getCaptureDate() {
        Date date = null;
        for(int i=0;i<NAME_POSSDATETIME.length;i++)
        {
    		String k = DIRECTORY_EXIF + sep + NAME_POSSDATETIME[i];
    		JpegMetadataTag jpegTag = (JpegMetadataTag) jpegTags.get(k);
    		if (jpegTag != null) {
    			String [] temp = jpegTag.getDescription().split(" ");
    			String [] tempDateArray = temp[0].split(":");
    			String [] tempTimeArray = temp[1].split(":");
    			int year = Integer.parseInt(tempDateArray[0]) - 1900; // Adjust for Date constructor
    			int month = Integer.parseInt(tempDateArray[1])-1;
    			int day = Integer.parseInt(tempDateArray[2]);
    			int hour = Integer.parseInt(tempTimeArray[0]);
    			int minute = Integer.parseInt(tempTimeArray[1]);
    			int seconds = Integer.parseInt(tempTimeArray[2]);
    			date = new Date(year, month, day, hour, minute, seconds);
    		}
        }
        return date;
	}
	
	/**
	 * Get EXIF Date/Time (time of capture of image) as String
	 * 
	 * @return Date/Time description ("YYYY:MM:DD HH:MM:SS")
	 * 
	 */
	public String getCaptureDateAsString() {
        for(int i=0;i<NAME_POSSDATETIME.length;i++)
        {
            String k = DIRECTORY_EXIF + sep + NAME_POSSDATETIME[i];
            if (jpegTags.containsKey(k)) {
                JpegMetadataTag jpegTag = (JpegMetadataTag) jpegTags.get(k);
                return jpegTag.getDescription();	
            }
        }
		return null;
	}
	
	/**
	 * Get EXIF width
	 * 
	 * @return width
	 * 
	 */
	public String getWidth() {
		String k = DIRECTORY_JPEG + sep + NAME_WIDTH;
		JpegMetadataTag jpegTag = (JpegMetadataTag) jpegTags.get(k);
		if (jpegTag != null) {
			return jpegTag.getDescription();
		}
		return null;
	}		
	
	/**
	 * Get EXIF height
	 * 
	 * @return height
	 * 
	 */
	public String getHeight() {
		String k = DIRECTORY_JPEG + sep + NAME_HEIGHT;
		JpegMetadataTag jpegTag = (JpegMetadataTag) jpegTags.get(k);
		if (jpegTag != null) {
			return jpegTag.getDescription();
		}
		return null;
	}

	/**
	 * Get all tags as ArrayList
	 * 
	 * @return tags
	 * 
	 */
	public ArrayList getTagList() {
		return jpegTagsList;
	}
}
