/*
 * Created on Jul 28, 2005
 *
 */
package gcr.mmm2.tag;

import gcr.mmm2.model.IContext;
import gcr.mmm2.model.IPerson;
import gcr.mmm2.model.IPhoto;
import gcr.mmm2.model.ObjectNotFoundException;
import gcr.mmm2.model.PhotoImpl;
import gcr.mmm2.model.ShareManager;
import gcr.mmm2.util.MultiValueDictionary;

import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Vector;

/**
 * Taggable multimedia files
 * 
 * TODO: should handle all media not just photos
 * 
 * @author Benjamin
 * 
 */
public class TaggableMedia extends Taggable {

    private static int cacheMiss = 0, cacheHit = 0;

    /**
     * List of people shared with
     */
    public static final int COMMUNITY_SHARES = 101;

    /**
     * Name of location
     */
    public static final int CONTENT_LOCATION = 202;

    /**
     * People in the photograph
     */
    public static final int CONTENT_PEOPLE_IN_MEDIA = 203;

    /**
     * Cell tower the photo was taken in
     */
    public static final int CONTEXT_CELL = 301;

    /**
     * GPS, either exact, estimated, or null of unknown
     */
    public static final int CONTEXT_GPS = 305;

    /**
     * List of people associated with bluetooth devices nearby
     */
    public static final int CONTEXT_PEOPLE_NEARBY = 306;

    /**
     * Who took the photograph
     */
    public static final int CONTEXT_PHOTOGRAPHER = 307;

    /**
     * Day of week in full text format
     */
    public static final int CONTEXT_TIME_DAYOFWEEK = 401;

    /**
     * Day of year (1-365)
     */
    public static final int CONTEXT_TIME_DAYOFYEAR = 402;

    /**
     * Month of year
     */
    public static final int CONTEXT_TIME_MONTH = 403;

    /**
     * Hour of day
     */
    public static final int CONTEXT_TIME_HOUR = 404;

    /**
     * Minute of Hour
     */
    public static final int CONTEXT_TIME_MINUTE = 405;

    /**
     * Month of year
     */
    public static final int CONTEXT_TIME_YEAR = 406;

    /**
     * Array of all types
     */
    public static final int[] MEDIA_TYPES = { COMMUNITY_SHARES,
            CONTEXT_PHOTOGRAPHER, CONTENT_LOCATION, CONTEXT_CELL, CONTEXT_GPS,
            CONTEXT_PEOPLE_NEARBY, CONTEXT_TIME_DAYOFWEEK,
            CONTEXT_TIME_YEAR, CONTEXT_TIME_MONTH, CONTEXT_TIME_DAYOFYEAR,
            CONTEXT_TIME_HOUR, CONTEXT_TIME_MINUTE
    // CONTEXT_WEATHER, CONTEXT_DAY_OF_YEAR, ,
    // CONTENT_PEOPLE_IN_MEDIA,
    };

    private static final MultiValueDictionary mvd = new MultiValueDictionary();

    protected static final Map registry = new Hashtable();

    private static final Hashtable tagMap = new Hashtable();

    private static final Hashtable writableMap = new Hashtable();

    static {
        for (int i = 0; i < getTypes().length; i++) {
            int type = getTypes()[i];
            tagMap.put("" + type, getTagDescription(type));
            if (isWritable(type)) {
                writableMap.put("" + type, getTagDescription(type));
            }
        }
    }

    public static String getTagDescription(int type) {
        switch (type) {
        case (COMMUNITY_SHARES):
            return "Photo Shared With";
        case (CONTEXT_PHOTOGRAPHER):
            return "Photographer";
        case (CONTENT_LOCATION):
            return "Location";
        case (CONTEXT_CELL):
            return "Cell Tower";
        case (CONTEXT_TIME_DAYOFWEEK):
            return "Day of Week";
        case (CONTEXT_TIME_DAYOFYEAR):
            return "Day of Year";
        case (CONTEXT_TIME_MONTH):
            return "Month";
        case (CONTEXT_TIME_YEAR):
            return "Year";
        case (CONTEXT_TIME_HOUR):
            return "Hour of Day";
        case (CONTEXT_TIME_MINUTE):
            return "Minute of Hour";
        case (CONTEXT_GPS):
            return "GPS Location";
        case (CONTEXT_PEOPLE_NEARBY):
            return "Sensed People";
        default:
            throw new IllegalArgumentException("Not yet implemented:" + type);
        }
    }

    /**
     * 
     * @param pid
     * @return Unique instance of TaggableMedia for that PhotoID
     * @throws ObjectNotFoundException
     */
    public static TaggableMedia getTaggablePhoto(int pid)
            throws ObjectNotFoundException {
        return getTaggablePhoto(new PhotoImpl(pid));
    }

    /**
     * Only one TaggableMedia per IPhoto is allowed.
     * 
     * @param p
     * @return the unique instance of TaggableMedia for the IPhoto
     */
    public static TaggableMedia getTaggablePhoto(IPhoto p) {
        Integer key = new Integer(p.getID());
        if (registry.containsKey(key))
            return (TaggableMedia) registry.get(key);

        TaggableMedia tp = new TaggableMedia(p);
        registry.put(key, tp);
        return tp;
    }

    public static Map getTagMap() {
        return tagMap;
    }

    public static int[] getTypes() {
        return MEDIA_TYPES;
    }

    public static Map getWritableTagMap() {
        return writableMap;
    }

    public static boolean isListType(int type) {

        switch (type) {
        case (COMMUNITY_SHARES):
        case (CONTENT_PEOPLE_IN_MEDIA):
        case (CONTEXT_PEOPLE_NEARBY):
        case (OTHER_FREE_TAG):
            return true;
        default:
            return false;
        }
    }

    public static boolean isWritable(int type) {

        switch (type) {
        case (CONTEXT_CELL):

        case (CONTEXT_GPS):
        case (CONTEXT_PEOPLE_NEARBY):
        case (CONTEXT_TIME_YEAR):
        case (CONTEXT_TIME_MONTH):
        case (CONTEXT_TIME_DAYOFWEEK):
        case (CONTEXT_TIME_DAYOFYEAR):
        case (CONTEXT_TIME_HOUR):
        case (CONTEXT_TIME_MINUTE):
        case (COMMUNITY_SHARES):
        case (CONTEXT_PHOTOGRAPHER):
            return false;
        case (CONTENT_LOCATION):
            return true;

        default:
            throw new IllegalArgumentException("Not yet implemented:" + type);
        }
    }

    private final SimpleDateFormat formatMinute = new SimpleDateFormat("mm");

    private final SimpleDateFormat formatYear = new SimpleDateFormat("yyyy");

    private final SimpleDateFormat formatMonth = new SimpleDateFormat("MMMM");

    private final SimpleDateFormat formatWeekday = new SimpleDateFormat("EEEE");

    private final SimpleDateFormat formatHour = new SimpleDateFormat("kk");

    private final SimpleDateFormat formatDayOfYear = new SimpleDateFormat("DD");

    private IPhoto photo;

    private int photo_id;

    private TaggableMedia(IPhoto p) {
        this.photo_id = p.getID();
        this.photo = p;
    }

    public void addOrSetTag(int type, String value)
            throws IllegalArgumentException {
        throw new IllegalArgumentException("Not yet implemented: set " + type
                + "=" + value);

    }

    /**
     * Gets every tag type, caching it locally.
     * 
     */
    public void cacheAll() {
        for (int i = 0; i < getTypes().length; i++) {
            this.getTag(getTypes()[i]);
        }
    }

    private String getCacheKey(int type) {
        return new StringBuffer().append(type).append(".").append(photo_id)
                .toString();
    }

    public int getMinuteDelta(TaggableMedia other) {
        int myTime = Integer.parseInt(getTag(CONTEXT_TIME_HOUR)[0]) * 60
                + Integer.parseInt(getTag(CONTEXT_TIME_MINUTE)[0]);
        int otherTime = Integer.parseInt(other.getTag(CONTEXT_TIME_HOUR)[0])
                * 60 + Integer.parseInt(other.getTag(CONTEXT_TIME_MINUTE)[0]);
        return Math.abs(myTime - otherTime);
    }

    /**
     * 
     * @param other
     *            Other TaggableMedia to find nearness to
     * @return Heuristically generated nearness from 0 to .8
     */
    public float getNearness(TaggableMedia other) {
        float nearness = 0.0f;
        final int timeDelta = getMinuteDelta(other);

        // Same Cell (0 to .1)
        int cell = getTagOverlap(CONTEXT_CELL, other);

        if (cell > 0 && timeDelta < 120) {
            nearness += .1;
            System.out.println("  CELL/TIME Overlap");
        }

        // Same people nearby (0 to .1)
        int nearby = getTagOverlap(CONTEXT_PEOPLE_NEARBY, other);
        if (nearby > 0) {
            nearness += .01 * Math.max(nearby, 10);
            System.out.println("  NEARBY PERSON Overlap:" + nearby);
        }

        return nearness;
    }

    /**
     * 
     * @return list of non empty tag types.
     * @see gcr.mmm2.tag.Taggable#getNonEmptyTagTypes()
     */
    public int[] getNonEmptyTagTypes() {
        final LinkedList nonEmpty = new LinkedList();
        for (int i = 0; i < getTypes().length; i++) {
            if (this.getTag(getTypes()[i]) != null
                    && this.getTag(getTypes()[i]).length > 0)
                nonEmpty.add(new Integer(getTypes()[i]));
        }
        int tags[] = new int[nonEmpty.size()];
        for (int i = 0; i < tags.length; i++) {
            tags[i] = ((Integer) nonEmpty.get(i)).intValue();
        }
        return tags;
    }

    public IPhoto getPhoto() {
        return this.photo;
    }

    public String[] getTag(int type) throws IllegalArgumentException {
        String tags[] = getTagCache(type);
        if (tags != null) {
            return tags;
        }

        tags = new String[1];

        IContext context = null;

        switch (type) {
        case (CONTEXT_CELL):
        case (CONTEXT_GPS):
        case (CONTEXT_PEOPLE_NEARBY):
        case (CONTENT_LOCATION):
            context = this.photo.getContext();
            if (context == null)
                return null;
            break;
        }

        switch (type) {

        case (COMMUNITY_SHARES):
            Iterator shares = ShareManager.getPhotoRecipients(this.photo,
                    this.photo.getOwner()).iterator();
            LinkedList sharePersons = new LinkedList();
            while (shares.hasNext()) {
                IPerson person = (IPerson) shares.next();
                sharePersons.add(person.getFirstName() + " "
                        + person.getLastName());
            }
            tags = (String[]) sharePersons.toArray(new String[0]);
            break;
        case (CONTENT_LOCATION):
            tags[0] = context.getLocationLabel();
            break;
        case (CONTEXT_CELL):
            tags[0] = context.getCellId();
            break;
        case (CONTEXT_GPS):
            tags[0] = context.getGpsLongitude() + ","
                    + context.getGpsLatitude();
            break;
        case (CONTEXT_PEOPLE_NEARBY):
            LinkedList persons = new LinkedList();
            Iterator people = context.listNearbyBluetoothPeople().iterator();

            while (people.hasNext()) {
                IPerson person = (IPerson) people.next();
                persons.add(person.toString());
            }
            tags = (String[]) persons.toArray(new String[0]);
            break;
        case (CONTEXT_TIME_DAYOFWEEK):
            tags[0] = formatWeekday.format(photo.getCaptureDate());
            break;
        case (CONTEXT_TIME_MONTH):
            tags[0] = formatMonth.format(photo.getCaptureDate());
            break;
        case (CONTEXT_TIME_DAYOFYEAR):
            tags[0] = formatDayOfYear.format(photo.getCaptureDate());
            break;
        case (CONTEXT_TIME_YEAR):
            tags[0] = formatYear.format(photo.getCaptureDate());
            break;
        case (CONTEXT_TIME_HOUR):
            tags[0] = formatHour.format(photo.getCaptureDate());
            break;
        case (CONTEXT_TIME_MINUTE):
            tags[0] = formatMinute.format(photo.getCaptureDate());
            break;

        case (CONTEXT_PHOTOGRAPHER):
            tags[0] = this.photo.getOwner().getFirstName() + " "
                    + this.photo.getOwner().getLastName();
            break;

        default:
            throw new IllegalArgumentException("Not yet implemented:" + type);
        }

        getTagCache(type, tags);
        return tags;
    }

    /**
     * 
     * @param type
     *            Tag Type
     * @return Values of tag for that item.
     * @throws IllegalArgumentException
     */
    private String[] getTagCache(int type) throws IllegalArgumentException {
        String key = getCacheKey(type);

        if ((cacheHit + cacheMiss) % 100 == 0) {
            System.out.println("TAG CACHE STATUS (hit/miss):" + cacheHit + "/"
                    + cacheMiss + ":" + key);
        }

        Vector vals = mvd.get(key);
        if (vals == null) {
            cacheMiss++;
            return null;
        }
        cacheHit++;
        return (String[]) vals.toArray(new String[0]);
    }

    /**
     * 
     * @param type
     *            Tag Type
     * @param values
     *            List of values to cache
     */
    private void getTagCache(int type, String[] values) {
        if (values == null)
            return;
        String key = getCacheKey(type);
        if (!isListType(type))
            mvd.remove(key);
        for (int i = 0; i < values.length; i++) {
            mvd.put(key, values[i]);
        }
    }

    public void removeTag(int type, String value)
            throws IllegalArgumentException {
        removeTagCache(type, value);
        throw new IllegalArgumentException("Not implemented:" + type + "="
                + value);
    }

    private void removeTagCache(int type, String value)
            throws IllegalArgumentException {
        if (!isWritable(type))
            throw new IllegalArgumentException("Non writable tag type:" + type);

        String key = getCacheKey(type);

        if (!isListType(type)) {
            mvd.remove(key);
        } else {
            mvd.remove(key, value);
        }
    }

}
