package gcr.mmm2.util;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

/**
 * The MultiValueDictionary class creates a dictionary object that can contain
 * multiple values for each key. The standard Java dictionary does not do this.
 * Values for a key are returned in a vector that contains all the values for
 * this key.
 * 
 * Internally the MultiValueDictionary is a Hashtable of Vectors. The key allows
 * in to the Hashtable and the values are contained in a Vector.
 *  
 */
public class MultiValueDictionary {
    private final Hashtable keys = new Hashtable();

    /**
     * Clears all of the keys and values from the dictionary
     */
    public synchronized void clear() {
        this.keys.clear();
    }

    /**
     * @param key
     * @return true if the dictionary has the key specified in it
     */
    public synchronized boolean containsKey(Object key) {
        return this.keys.containsKey(key);
    }

    /**
     * @param key
     * @param value
     * @return true if the dictionary has the key AND value specified in it
     */
    public synchronized boolean containsKeyValue(Object key, Object value) {
        if (this.keys.containsKey(key) == false) {
            return false;
        }

        Vector tempVect = (Vector) this.keys.get(key);
        return tempVect.contains(value);
    }

    /**
     * Sees if the dictionary has the specified value in ANY key.
     * 
     * @param value
     *            The value that must appear anywhere in the dictionary as a
     *            value
     * @return true or false
     */
    public synchronized boolean containsValue(Object value) {
        Vector tempVect;
        Enumeration e = this.keys.elements();
        while (e.hasMoreElements()) {
            tempVect = (Vector) e.nextElement();
            if (tempVect.contains(value))
                return true;
        }
        return false;

    }

    /**
     * This checks to see if the two MultiValueDictionaries contain an equal
     * hashtable. If so, then they're equal. Probably won't be used much.
     * (non-Javadoc)
     * 
     * @see java.lang.Object#equals(java.lang.Object)
     */
    public synchronized boolean equals(Object o) {
        if (o == null)
            return false;
        if (!o.getClass().equals(this.getClass()))
            return false;
        MultiValueDictionary dict = (MultiValueDictionary) o;
        return dict.getHashtable().equals(this.getHashtable());
    }

    /**
     * Returns a vector containing all the values that this key is mapped to.
     * Returns null if this key doesn't exist.
     * 
     * @param key
     *            the key you wish to obtain values for
     * @return vector of values
     */
    public synchronized Vector get(Object key) {
        if (!this.containsKey(key))
            return null;
        return (Vector) this.keys.get(key);
    }

    /**
     * Puts this value into the specified key. If the value already exists in
     * this key it does not insert duplicates.
     * 
     * @param key
     *            the key
     * @param value
     *            the value
     */
    public synchronized void put(Object key, Object value) {
        // if there are no values for this, then add a vector
        if (!this.containsKey(key)) {
            this.keys.put(key, new Vector());
        }

        Vector tempVec = (Vector) this.keys.get(key);
        if (tempVec.contains(value))
            return;
        tempVec.add(value);
    }

    /**
     * Removes this key from the dictionary. WARNING: removes all values
     * associated with this key.
     * 
     * @param key
     *            the key to remove
     * @return return the vector of the values for the removed key
     */
    public synchronized Vector remove(Object key) {
        return (Vector) this.keys.remove(key);
    }

    /**
     * Removes the object with the specified key AND value from the dictionary.
     * 
     * @param key
     *            The specific key to remove
     * @param value
     *            The specific value to remove from the key
     */
    public synchronized void remove(Object key, Object value) {
        if (!this.containsKey(key))
            return;
        Vector tempVect = (Vector) this.keys.get(key);
        tempVect.remove(value);
    }

    /**
     * Get the count of keys in the dictionary
     * 
     * @return the number of keys
     */
    public synchronized int keyCount() {
        return this.keys.size();
    }

    /**
     * The count of the number of values for all keys
     * 
     * @return an integer for number of values
     */
    public synchronized int valueCount() {
        Vector tempVect;
        int totalSize = 0;
        Enumeration e = this.keys.elements();
        while (e.hasMoreElements()) {
            tempVect = (Vector) e.nextElement();
            totalSize += tempVect.size();
        }
        return totalSize;
    }

    /**
     * returns a hashtable containing keys and vectors as values. Each vector
     * lists the values for that key.
     * 
     * @return the hashtable of vectors
     */
    public synchronized Hashtable getHashtable() {
        return this.keys;
    }

    /**
     * returns the keys
     * 
     * @return Enumeration of the keys
     */
    public synchronized Enumeration keys() {
        return this.keys.keys();
    }
}
